@@ -18,7 +18,7 @@
  */
 
 
-#include <malloc.h>
+#include <stdlib.h>
 #include <string.h>
 #include <math.h>	// fabs, hypotf
 #include <float.h>	// FLT_EPSILON
@@ -35,8 +35,7 @@ fsk_plan_new(
 	float		sample_rate,
     	float		f_mark,
     	float		f_space,
-	float		filter_bw,
-	unsigned int	n_data_bits
+	float		filter_bw
 	)
 {
     fsk_plan *fskp = malloc(sizeof(fsk_plan));
@@ -46,10 +45,6 @@ fsk_plan_new(
     fskp->sample_rate = sample_rate;
     fskp->f_mark = f_mark;
     fskp->f_space = f_space;
-    fskp->n_data_bits = n_data_bits;
-
-    /* 1 prev_stop + n_data_bits + 1 start + 1 stop == n_data_bits + 3 */
-    fskp->n_frame_bits = fskp->n_data_bits + 3;
 
 #ifdef USE_FFT
     fskp->band_width = filter_bw;
@@ -179,27 +174,12 @@ fsk_bit_analyze( fsk_plan *fskp, float *samples, unsigned int bit_nsamples,
 }
 
 
-/* returns confidence value [0.0 to 1.0] */
+/* returns confidence value [0.0 to INFINITY] */
 static float
 fsk_frame_analyze( fsk_plan *fskp, float *samples, float samples_per_bit,
-	unsigned int *bits_outp )
+	int n_bits, const char *expect_bits_string,
+	unsigned int *bits_outp, float *ampl_outp )
 {
-    int n_bits = fskp->n_frame_bits;
-//    char *expect_bit_string = "10dddddddd1";
-    char expect_bit_string[32];
-    memset(expect_bit_string, 'd', 32);
-    expect_bit_string[0] = '1';
-    expect_bit_string[1] = '0';
-    expect_bit_string[fskp->n_data_bits + 2] = '1';
-
-    // example...
-    // 0123456789A
-    // isddddddddp	i == idle bit (a.k.a. prev_stop bit)
-    //			s == start bit
-    //			d == data bits
-    //			p == stop bit
-    // MSddddddddM  <-- expected mark/space framing pattern
-
     unsigned int bit_nsamples = (float)(samples_per_bit + 0.5);
 
     unsigned int	bit_values[32];
@@ -208,7 +188,12 @@ fsk_frame_analyze( fsk_plan *fskp, float *samples, float samples_per_bit,
     unsigned int	bit_begin_sample;
     int			bitnum;
 
-    char *expect_bits = expect_bit_string;
+// various deprecated noise limiter schemes:
+//#define FSK_MIN_BIT_SNR 1.4
+//#define FSK_MIN_MAGNITUDE 0.10
+//#define FSK_AVOID_TRANSIENTS	0.7
+
+    const char *expect_bits = expect_bits_string;
 
     /* pass #1 - process and check only the "required" (1/0) expect_bits */
     for ( bitnum=0; bitnum<n_bits; bitnum++ ) {
@@ -226,8 +211,11 @@ fsk_frame_analyze( fsk_plan *fskp, float *samples, float samples_per_bit,
 	if ( (expect_bits[bitnum] - '0') != bit_values[bitnum] )
 	    return 0.0; /* does not match expected; abort frame analysis. */
 
-	// FSK_MIN_MAGNITUDE serves as a noise limiter
-# define FSK_MIN_MAGNITUDE 0.10
+#ifdef FSK_MIN_BIT_SNR
+	float bit_snr = bit_sig_mags[bitnum] / bit_noise_mags[bitnum];
+	if ( bit_snr < FSK_MIN_BIT_SNR )
+	    return 0.0;
+#endif
 
 # ifdef FSK_MIN_MAGNITUDE
 	// Performance hack: reject frame early if sig mag isn't even half
@@ -235,20 +223,20 @@ fsk_frame_analyze( fsk_plan *fskp, float *samples, float samples_per_bit,
 	if ( bit_sig_mags[bitnum] < FSK_MIN_MAGNITUDE/2.0 )
 	    return 0.0; // too weak; abort frame analysis
 # endif
-
     }
 
-// Note: CONFIDENCE_ALGO 3 does not need AVOID_TRANSIENTS
-#define AVOID_TRANSIENTS	0.7
-//
-#ifdef AVOID_TRANSIENTS
+#ifdef FSK_AVOID_TRANSIENTS
+    // FIXME: fsk_frame_analyze shouldn't care about start/stop bits,
+    // and this really is only correct for "10dd..dd1" format frames anyway:
+    // FIXME: this is totally defective, if the checked bits weren't
+    // even calculated in pass #1 (e.g. if there are no pass #1 expect bits).
     /* Compare strength of stop bit and start bit, to avoid detecting
      * a transient as a start bit, as often results in a single false
      * character when the mark "leader" tone begins.  Require that the
      * diff between start bit and stop bit strength not be "large". */
-    float s_mag = bit_sig_mags[1];
-    float p_mag = bit_sig_mags[fskp->n_data_bits + 2];
-    if ( fabs(s_mag-p_mag) > (s_mag * AVOID_TRANSIENTS) ) {
+    float s_mag = bit_sig_mags[1]; // start bit
+    float p_mag = bit_sig_mags[n_bits-1]; // stop bit
+    if ( fabs(s_mag-p_mag) > (s_mag * FSK_AVOID_TRANSIENTS) ) {
 	debug_log(" avoid transient\n");
 	return 0.0;
     }
@@ -264,17 +252,24 @@ fsk_frame_analyze( fsk_plan *fskp, float *samples, float samples_per_bit,
 		&bit_values[bitnum],
 		&bit_sig_mags[bitnum],
 		&bit_noise_mags[bitnum]);
+
+#ifdef FSK_MIN_BIT_SNR
+	float bit_snr = bit_sig_mags[bitnum] / bit_noise_mags[bitnum];
+	if ( bit_snr < FSK_MIN_BIT_SNR )
+	    return 0.0;
+#endif
     }
 
 
-#define CONFIDENCE_ALGO	5
+//#define CONFIDENCE_ALGO 	5
+#define CONFIDENCE_ALGO 	6
 
     float confidence;
 
-#if CONFIDENCE_ALGO == 5
+#if CONFIDENCE_ALGO == 5 || CONFIDENCE_ALGO == 6
 
     float total_bit_sig = 0.0, total_bit_noise = 0.0;
-    for ( bitnum=0; bitnum<(n_bits-1); bitnum++ ) {
+    for ( bitnum=0; bitnum<n_bits; bitnum++ ) {
 	// Deal with floating point data type quantization noise...
 	// If total_bit_noise <= FLT_EPSILON, then assume it to be 0.0,
 	// so that we end up with snr==inf.
@@ -286,23 +281,49 @@ fsk_frame_analyze( fsk_plan *fskp, float *samples, float samples_per_bit,
     // Compute the "frame SNR"
     float snr = total_bit_sig / total_bit_noise;
 
-    float avg_bit_sig   = total_bit_sig / (n_bits-1);
+    // Compute avg bit sig and noise magnitudes
+    float avg_bit_sig   = total_bit_sig / n_bits;
+
+#if CONFIDENCE_ALGO == 6
+    // Compute average "divergence": bit_mag_divergence / other_bits_mag
+    float divergence = 0.0;
+    for ( bitnum=0; bitnum<n_bits; bitnum++ ) {
+	float avg_bit_sig_other = (total_bit_sig - fabs(bit_sig_mags[bitnum]))
+					/ (n_bits - 1);
+	divergence += fabs(bit_sig_mags[bitnum] - avg_bit_sig_other)
+					/ avg_bit_sig_other;
+    }
+    divergence *= 2;
+    divergence /= n_bits;
+#endif
+
 #ifdef FSK_DEBUG
-    float avg_bit_noise = total_bit_noise / (n_bits-1);
-    debug_log("    snr=%.6f avg{ bit_sig=%.6f bit_noise=%.6f (%s) }\n",
+    float avg_bit_noise = total_bit_noise / n_bits;
+    debug_log("    divg=%.3f snr=%.3f avg{bit_sig=%.3f bit_noise=%.3f(%s)}\n",
+# if CONFIDENCE_ALGO == 6
+	    divergence,
+# else
+	    0.0,
+# endif
 	    snr, avg_bit_sig, avg_bit_noise,
 	    avg_bit_noise == 0.0 ? "zero" : "non-zero"
 	    );
 #endif
 
 # ifdef FSK_MIN_MAGNITUDE
-    // noise limiter
     if ( avg_bit_sig < FSK_MIN_MAGNITUDE )
 	return 0.0; // too weak; reject frame
 # endif
 
+#if CONFIDENCE_ALGO == 6
+    // Frame confidence is the frame ( SNR * consistency )
+    confidence = snr * (1.0 - divergence);
+#else
     // Frame confidence is the frame SNR
     confidence = snr;
+#endif
+
+    *ampl_outp = avg_bit_sig;
 
 #elif CONFIDENCE_ALGO == 4
 
@@ -397,12 +418,14 @@ fsk_frame_analyze( fsk_plan *fskp, float *samples, float samples_per_bit,
 #endif /* CONFIDENCE_ALGO */
 
 
+    // least significant bit first ... reverse the bits as we place them
+    // into the bits_outp word.
     *bits_outp = 0;
     for ( bitnum=0; bitnum<n_bits; bitnum++ )
 	*bits_outp |= bit_values[bitnum] << bitnum;
 
-    debug_log("    frame confidence (algo #%u) = %f\n",
-	    CONFIDENCE_ALGO, confidence);
+    debug_log("    frame algo=%u confidence=%f ampl=%f\n",
+	    CONFIDENCE_ALGO, confidence, *ampl_outp);
     return confidence;
 }
 
@@ -413,16 +436,20 @@ fsk_find_frame( fsk_plan *fskp, float *samples, unsigned int frame_nsamples,
 	unsigned int try_max_nsamples,
 	unsigned int try_step_nsamples,
 	float try_confidence_search_limit,
+	const char *expect_bits_string,
 	unsigned int *bits_outp,
+	float *ampl_outp,
 	unsigned int *frame_start_outp
 	)
 {
-    float samples_per_bit = (float)frame_nsamples / fskp->n_frame_bits;
+    int expect_n_bits = strlen(expect_bits_string);
+
+    float samples_per_bit = (float)frame_nsamples / expect_n_bits;
 
     // try_step_nsamples = 1;	// pedantic TEST
 
     unsigned int best_t = 0;
-    float best_c = 0.0;
+    float best_c = 0.0, best_a = 0.0;
     unsigned int best_bits = 0;
     
     // Scan the frame positions starting with the one try_first_sample,
@@ -438,13 +465,16 @@ fsk_find_frame( fsk_plan *fskp, float *samples, unsigned int frame_nsamples,
 	if ( t < 0 )
 	    continue;
 
-	float c;
+	float c, ampl_out;
 	unsigned int bits_out = 0;
 	debug_log("try fsk_frame_analyze at t=%d\n", t);
-	c = fsk_frame_analyze(fskp, samples+t, samples_per_bit, &bits_out);
+	c = fsk_frame_analyze(fskp, samples+t, samples_per_bit,
+			expect_n_bits, expect_bits_string,
+			&bits_out, &ampl_out);
 	if ( best_c < c ) {
 	    best_t = t;
 	    best_c = c;
+	    best_a = ampl_out;
 	    best_bits = bits_out;
 	    // If we find a frame with confidence > try_confidence_search_limit
 	    // quit searching.
@@ -454,6 +484,7 @@ fsk_find_frame( fsk_plan *fskp, float *samples, unsigned int frame_nsamples,
     }
 
     *bits_outp = best_bits;
+    *ampl_outp = best_a;
     *frame_start_outp = best_t;
 
     float confidence = best_c;
@@ -463,11 +494,15 @@ fsk_find_frame( fsk_plan *fskp, float *samples, unsigned int frame_nsamples,
 
     // FIXME? hardcoded chop off framing bits for debug
 #ifdef FSK_DEBUG
-    unsigned char bitchar = ( best_bits >> 2 ) & 0xFF;
-    debug_log("FSK_FRAME datum='%c' (0x%02x)   c=%f  t=%d\n",
+    unsigned char bitchar = ( *bits_outp >> 2 ) & 0xFF;
+
+    debug_log("FSK_FRAME bits='");
+    for ( j=0; j<expect_n_bits; j++ )
+	debug_log("%c", ( *bits_outp >> j ) & 1 ? '1' : '0' );
+    debug_log("' datum='%c' (0x%02x)   c=%f  a=%f  t=%d\n",
 	    isprint(bitchar)||isspace(bitchar) ? bitchar : '.',
 	    bitchar,
-	    confidence, best_t);
+	    confidence, best_a, best_t);
 #endif
 
     return confidence;
