@@ -21,11 +21,13 @@
 
 
 #include <getopt.h>
+#include <unistd.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
 #include <math.h>
+#include <float.h>
 #include <assert.h>
 #include <signal.h>
 #include <sys/time.h>
@@ -38,59 +40,10 @@
 
 #include "simpleaudio.h"
 #include "fsk.h"
-#include "baudot.h"
+#include "databits.h"
 
 char *program_name = "";
 
-
-/*
- * ASCII 8-bit data framebits decoder/encoder (passthrough)
- */
-
-/* returns the number of datawords stuffed into *databits_outp */
-int
-framebits_encode_ascii8( unsigned int *databits_outp, char char_out )
-{
-    *databits_outp = char_out;
-    return 1;
-}
-
-/* returns nbytes decoded */
-static unsigned int
-framebits_decode_ascii8( char *dataout_p, unsigned int dataout_size,
-	unsigned int bits )
-{
-    if ( dataout_p == NULL )	// frame processor reset: noop
-	return 0;
-    assert( (bits & ~0xFF) == 0 );
-    assert( dataout_size >= 1);
-    *dataout_p = bits;
-    return 1;
-}
-
-
-/*
- * Baudot 5-bit data framebits decoder/encoder
- */
-
-#define framebits_encode_baudot baudot_encode
-
-/* returns nbytes decoded */
-static unsigned int
-framebits_decode_baudot( char *dataout_p, unsigned int dataout_size,
-	unsigned int bits )
-{
-    if ( dataout_p == NULL ) {	// frame processor reset: reset Baudot state
-	    baudot_reset();
-	    return 0;
-    }
-    assert( (bits & ~0x1F) == 0 );
-    assert( dataout_size >= 1);
-    return baudot_decode(dataout_p, bits);
-}
-
-
-
 int		tx_transmitting = 0;
 int		tx_leader_bits_len = 2;
 int		tx_trailer_bits_len = 2;
@@ -126,8 +79,8 @@ static void fsk_transmit_stdin(
 	float bfsk_mark_f,
 	float bfsk_space_f,
 	int n_data_bits,
-	float bfsk_txstopbits,
-	int (*framebits_encoder)( unsigned int *databits_outp, char char_out )
+	float bfsk_nstopbits,
+	databits_encoder encode
 	)
 {
     size_t sample_rate = simpleaudio_get_rate(sa_out);
@@ -156,7 +109,7 @@ static void fsk_transmit_stdin(
 	// fprintf(stderr, "<c=%d>", c);
 	unsigned int nwords;
 	unsigned int bits[2];
-	nwords = framebits_encoder(bits, c);
+	nwords = encode(bits, c);
 
 	if ( !tx_transmitting )
 	{
@@ -175,7 +128,7 @@ static void fsk_transmit_stdin(
 		simpleaudio_tone(sa_out, tone_freq, bit_nsamples);
 	    }
 	    simpleaudio_tone(sa_out, bfsk_mark_f,
-				bit_nsamples * bfsk_txstopbits);	// stop
+				bit_nsamples * bfsk_nstopbits);	// stop
 	}
 
 	if ( tx_interactive )
@@ -196,24 +149,31 @@ static void
 report_no_carrier( fsk_plan *fskp,
 	unsigned int sample_rate,
 	float bfsk_data_rate,
-	float nsamples_per_bit,
+	float frame_n_bits,
 	unsigned int nframes_decoded,
 	size_t carrier_nsamples,
-	float confidence_total )
+	float confidence_total,
+	float amplitude_total )
 {
-    unsigned long long nbits_total = nframes_decoded * (fskp->n_data_bits+2);
+    float nbits_decoded = nframes_decoded * frame_n_bits;
 #if 0
     fprintf(stderr, "nframes_decoded=%u\n", nframes_decoded);
-    fprintf(stderr, "nbits_total=%llu\n", nbits_total);
+    fprintf(stderr, "nbits_decoded=%f\n", nbits_decoded);
     fprintf(stderr, "carrier_nsamples=%lu\n", carrier_nsamples);
-    fprintf(stderr, "nsamples_per_bit=%f\n", nsamples_per_bit);
 #endif
-    float throughput_rate = nbits_total * sample_rate / (float)carrier_nsamples;
-    fprintf(stderr, "### NOCARRIER ndata=%u confidence=%.3f throughput=%.2f",
+    float throughput_rate =
+		nbits_decoded * sample_rate / (float)carrier_nsamples;
+    fprintf(stderr, "\n### NOCARRIER ndata=%u confidence=%.3f ampl=%.3f bps=%.2f",
 	    nframes_decoded,
 	    confidence_total / nframes_decoded,
+	    amplitude_total / nframes_decoded,
 	    throughput_rate);
-    if ( (size_t)(nbits_total * nsamples_per_bit + 0.5) == carrier_nsamples ) {
+#if 0
+    fprintf(stderr, " bits*sr=%llu rate*nsamp=%llu",
+	    (unsigned long long)(nbits_decoded * sample_rate + 0.5),
+	    (unsigned long long)(bfsk_data_rate * carrier_nsamples) );
+#endif
+    if ( (unsigned long long)(nbits_decoded * sample_rate + 0.5) == (unsigned long long)(bfsk_data_rate * carrier_nsamples) ) {
 	fprintf(stderr, " (rate perfect) ###\n");
     } else {
 	float throughput_skew = (throughput_rate - bfsk_data_rate)
@@ -256,9 +216,9 @@ benchmarks()
 
 
     // enable the sine wave LUT
-    simpleaudio_tone_init(1024);
+    simpleaudio_tone_init(1024, 1.0);
 
-    sa_out = simpleaudio_open_stream(backend, SA_STREAM_PLAYBACK,
+    sa_out = simpleaudio_open_stream(backend, NULL, SA_STREAM_PLAYBACK,
 			SA_SAMPLE_FORMAT_S16, sample_rate, 1,
 			program_name, "generate-tones-lut1024-S16-mono");
     if ( ! sa_out )
@@ -266,7 +226,7 @@ benchmarks()
     generate_test_tones(sa_out, 10);
     simpleaudio_close(sa_out);
 
-    sa_out = simpleaudio_open_stream(backend, SA_STREAM_PLAYBACK,
+    sa_out = simpleaudio_open_stream(backend, NULL, SA_STREAM_PLAYBACK,
 			SA_SAMPLE_FORMAT_FLOAT, sample_rate, 1,
 			program_name, "generate-tones-lut1024-FLOAT-mono");
     if ( ! sa_out )
@@ -276,9 +236,9 @@ benchmarks()
 
 
     // disable the sine wave LUT
-    simpleaudio_tone_init(0);
+    simpleaudio_tone_init(0, 1.0);
 
-    sa_out = simpleaudio_open_stream(backend, SA_STREAM_PLAYBACK,
+    sa_out = simpleaudio_open_stream(backend, NULL, SA_STREAM_PLAYBACK,
 			SA_SAMPLE_FORMAT_S16, sample_rate, 1,
 			program_name, "generate-tones-nolut-S16-mono");
     if ( ! sa_out )
@@ -286,7 +246,7 @@ benchmarks()
     generate_test_tones(sa_out, 10);
     simpleaudio_close(sa_out);
 
-    sa_out = simpleaudio_open_stream(backend, SA_STREAM_PLAYBACK,
+    sa_out = simpleaudio_open_stream(backend, NULL, SA_STREAM_PLAYBACK,
 			SA_SAMPLE_FORMAT_FLOAT, sample_rate, 1,
 			program_name, "generate-tones-nolut-FLOAT-mono");
     if ( ! sa_out )
@@ -299,6 +259,15 @@ benchmarks()
 }
 
 
+static int rx_stop = 0;
+
+void
+rx_stop_sighandler( int sig )
+{
+    rx_stop = 1;
+}
+
+
 void
 version()
 {
@@ -321,27 +290,36 @@ usage()
     "		    -r, --rx, --receive,  --read     (default)\n"
     "		[options]\n"
     "		    -a, --auto-carrier\n"
-    "		    -c, --confidence {min-snr-threshold}\n"
-    "		    -l, --limit {max-snr-search-limit}\n"
+    "		    -i, --inverted\n"
+    "		    -c, --confidence {min-confidence-threshold}\n"
+    "		    -l, --limit {max-confidence-search-limit}\n"
     "		    -8, --ascii		ASCII  8-N-1\n"
     "		    -5, --baudot	Baudot 5-N-1\n"
     "		    -f, --file {filename.flac}\n"
     "		    -b, --bandwidth {rx_bandwidth}\n"
+    "		    -v, --volume {amplitude or 'E'}\n"
     "		    -M, --mark {mark_freq}\n"
     "		    -S, --space {space_freq}\n"
-    "		    -T, --txstopbits {m.n}\n"
+    "		    --startbits {n}\n"
+    "		    --stopbits {n.n}\n"
+    "		    --sync-byte {0xXX}\n"
     "		    -q, --quiet\n"
     "		    -R, --samplerate {rate}\n"
     "		    -V, --version\n"
-    "		    -A, --alsa\n"
+    "		    -A, --alsa[=plughw:X,Y]\n"
     "		    --lut={tx_sin_table_len}\n"
     "		    --float-samples\n"
+    "		    --rx-one\n"
     "		    --benchmarks\n"
+    "		    --binary-output\n"
+    "		    --print-filter\n"
     "		{baudmode}\n"
-    "		    1200       Bell202  1200 bps --ascii\n"
-    "		     300       Bell103   300 bps --ascii\n"
-    "	    any_number_N       Bell103     N bps --ascii\n"
-    "		    rtty       RTTY    45.45 bps --baudot\n"
+    "	    any_number_N       Bell-like      N bps --ascii\n"
+    "		    1200       Bell202     1200 bps --ascii\n"
+    "		     300       Bell103      300 bps --ascii\n"
+    "		    rtty       RTTY       45.45 bps --baudot --stopbits=1.5\n"
+    "		    same       NOAA SAME 520.83 bps --sync-byte=0xAB ...\n"
+    "		callerid       Bell202 CID 1200 bps\n"
     );
     exit(1);
 }
@@ -352,10 +330,15 @@ main( int argc, char*argv[] )
     char *modem_mode = NULL;
     int TX_mode = -1;
     int quiet_mode = 0;
+    int output_print_filter = 0;
     float band_width = 0;
     unsigned int bfsk_mark_f = 0;
     unsigned int bfsk_space_f = 0;
-    float bfsk_txstopbits = 0;
+    unsigned int bfsk_inverted_freqs = 0;
+    int bfsk_nstartbits = -1;	// actually only supports 0 or 1 startbit
+    float bfsk_nstopbits = -1;
+    unsigned int bfsk_sync_on_data = 0;
+    unsigned int bfsk_sync_byte = -1;
     unsigned int bfsk_n_data_bits = 0;
     int autodetect_shift;
     char *filename = NULL;
@@ -364,25 +347,33 @@ main( int argc, char*argv[] )
 
     // fsk_confidence_threshold : signal-to-noise squelch control
     //
-    // The minimum SNR confidence level seen as "a signal".
-    float fsk_confidence_threshold = 2.0;
+    // The minimum SNR-ish confidence level seen as "a signal".
+    float fsk_confidence_threshold = 1.5;
 
     // fsk_confidence_search_limit : performance vs. quality
     //
-    // If we find a frame with SNR confidence > confidence_search_limit,
+    // If we find a frame with confidence > confidence_search_limit,
     // quit searching for a better frame.  confidence_search_limit has a
     // dramatic effect on peformance (high value yields low performance, but
-    // higher decode quality, for noisy or hard-to-discern signals (Bell 103).
+    // higher decode quality, for noisy or hard-to-discern signals (Bell 103,
+    // or skewed rates).
     float fsk_confidence_search_limit = 2.3f;
     // float fsk_confidence_search_limit = INFINITY;  /* for test */
 
     sa_backend_t sa_backend = SA_BACKEND_SYSDEFAULT;
+    char *sa_backend_device = NULL;
     sa_format_t sample_format = SA_SAMPLE_FORMAT_S16;
     unsigned int sample_rate = 48000;
     unsigned int nchannels = 1; // FIXME: only works with one channel
 
+    float tx_amplitude = 1.0;
     unsigned int tx_sin_table_len = 4096;
 
+    unsigned int rx_one = 0;
+    float rxnoise_factor = 0.0;
+
+    int output_mode_binary = 0;
+
     /* validate the default system audio mechanism */
 #if !(USE_PULSEAUDIO || USE_ALSA)
 # define _MINIMODEM_NO_SYSTEM_AUDIO
@@ -402,9 +393,16 @@ main( int argc, char*argv[] )
     
     enum {
 	MINIMODEM_OPT_UNUSED=256,	// placeholder
+	MINIMODEM_OPT_STARTBITS,
+	MINIMODEM_OPT_STOPBITS,
+	MINIMODEM_OPT_SYNC_BYTE,
 	MINIMODEM_OPT_LUT,
 	MINIMODEM_OPT_FLOAT_SAMPLES,
+	MINIMODEM_OPT_RX_ONE,
 	MINIMODEM_OPT_BENCHMARKS,
+	MINIMODEM_OPT_BINARY_OUTPUT,
+	MINIMODEM_OPT_PRINT_FILTER,
+	MINIMODEM_OPT_XRXNOISE,
     };
 
     while ( 1 ) {
@@ -419,22 +417,30 @@ main( int argc, char*argv[] )
 	    { "confidence",	1, 0, 'c' },
 	    { "limit",		1, 0, 'l' },
 	    { "auto-carrier",	0, 0, 'a' },
+	    { "inverted",	0, 0, 'i' },
 	    { "ascii",		0, 0, '8' },
 	    { "baudot",		0, 0, '5' },
 	    { "file",		1, 0, 'f' },
 	    { "bandwidth",	1, 0, 'b' },
+	    { "volume",		1, 0, 'v' },
 	    { "mark",		1, 0, 'M' },
 	    { "space",		1, 0, 'S' },
-	    { "txstopbits",	1, 0, 'T' },
+	    { "startbits",	1, 0, MINIMODEM_OPT_STARTBITS },
+	    { "stopbits",	1, 0, MINIMODEM_OPT_STOPBITS },
+	    { "sync-byte",	1, 0, MINIMODEM_OPT_SYNC_BYTE },
 	    { "quiet",		0, 0, 'q' },
-	    { "alsa",		0, 0, 'A' },
+	    { "alsa",		2, 0, 'A' },
 	    { "samplerate",	1, 0, 'R' },
 	    { "lut",		1, 0, MINIMODEM_OPT_LUT },
 	    { "float-samples",	0, 0, MINIMODEM_OPT_FLOAT_SAMPLES },
+	    { "rx-one",		0, 0, MINIMODEM_OPT_RX_ONE },
 	    { "benchmarks",	0, 0, MINIMODEM_OPT_BENCHMARKS },
+	    { "binary-output",	0, 0, MINIMODEM_OPT_BINARY_OUTPUT },
+	    { "print-filter",	0, 0, MINIMODEM_OPT_PRINT_FILTER },
+	    { "Xrxnoise",	1, 0, MINIMODEM_OPT_XRXNOISE },
 	    { 0 }
 	};
-	c = getopt_long(argc, argv, "Vtrc:a85f:b:M:S:T:qAR:",
+	c = getopt_long(argc, argv, "Vtrc:l:ai85f:b:v:M:S:T:qA::R:",
 		long_options, &option_index);
 	if ( c == -1 )
 	    break;
@@ -461,6 +467,9 @@ main( int argc, char*argv[] )
 	    case 'a':
 			carrier_autodetect_threshold = 0.001;
 			break;
+	    case 'i':
+			bfsk_inverted_freqs = 1;
+			break;
 	    case 'f':
 			filename = optarg;
 			break;
@@ -474,6 +483,13 @@ main( int argc, char*argv[] )
 			band_width = atof(optarg);
 			assert( band_width != 0 );
 			break;
+	    case 'v':
+			if ( optarg[0] == 'E' )
+			    tx_amplitude = FLT_EPSILON;
+			else
+			    tx_amplitude = atof(optarg);
+			assert( tx_amplitude > 0.0 );
+			break;
 	    case 'M':
 			bfsk_mark_f = atoi(optarg);
 			assert( bfsk_mark_f > 0 );
@@ -482,9 +498,17 @@ main( int argc, char*argv[] )
 			bfsk_space_f = atoi(optarg);
 			assert( bfsk_space_f > 0 );
 			break;
-	    case 'T':
-			bfsk_txstopbits = atof(optarg);
-			assert( bfsk_txstopbits > 0 );
+	    case MINIMODEM_OPT_STARTBITS:
+			bfsk_nstartbits = atoi(optarg);
+			assert( bfsk_nstartbits == 0 || bfsk_nstartbits == 1 );
+			break;
+	    case MINIMODEM_OPT_STOPBITS:
+			bfsk_nstopbits = atof(optarg);
+			assert( bfsk_nstopbits >= 0 );
+			break;
+	    case MINIMODEM_OPT_SYNC_BYTE:
+			bfsk_sync_on_data = 1;
+			bfsk_sync_byte = strtol(optarg, NULL, 0);
 			break;
 	    case 'q':
 			quiet_mode = 1;
@@ -496,6 +520,8 @@ main( int argc, char*argv[] )
 	    case 'A':
 #if USE_ALSA
 			sa_backend = SA_BACKEND_ALSA;
+			if ( optarg )
+			    sa_backend_device = optarg;
 #else
 			fprintf(stderr, "E: This build of minimodem was configured without alsa support.\n");
 			exit(1);
@@ -507,10 +533,22 @@ main( int argc, char*argv[] )
 	    case MINIMODEM_OPT_FLOAT_SAMPLES:
 			sample_format = SA_SAMPLE_FORMAT_FLOAT;
 			break;
+	    case MINIMODEM_OPT_RX_ONE:
+			rx_one = 1;
+			break;
 	    case MINIMODEM_OPT_BENCHMARKS:
 			benchmarks();
 			exit(0);
 			break;
+	    case MINIMODEM_OPT_BINARY_OUTPUT:
+			output_mode_binary = 1;
+			break;
+	    case MINIMODEM_OPT_PRINT_FILTER:
+			output_print_filter = 1;
+			break;
+	    case MINIMODEM_OPT_XRXNOISE:
+			rxnoise_factor = atof(optarg);
+			break;
 	    default:
 			usage();
 	}
@@ -552,17 +590,39 @@ main( int argc, char*argv[] )
 
 
     float	bfsk_data_rate = 0.0;
-    int (*bfsk_framebits_encode)( unsigned int *databits_outp, char char_out );
+    databits_encoder	*bfsk_databits_encode;
+    databits_decoder	*bfsk_databits_decode;
 
-    unsigned int (*bfsk_framebits_decode)( char *dataout_p, unsigned int dataout_size,
-					unsigned int bits );
+    bfsk_databits_decode = databits_decode_ascii8;
+    bfsk_databits_encode = databits_encode_ascii8;
 
     if ( strncasecmp(modem_mode, "rtty",5)==0 ) {
+	bfsk_databits_decode = databits_decode_baudot;
+	bfsk_databits_encode = databits_encode_baudot;
 	bfsk_data_rate = 45.45;
 	if ( bfsk_n_data_bits == 0 )
 	    bfsk_n_data_bits = 5;
-	if ( bfsk_txstopbits == 0 )
-	    bfsk_txstopbits = 1.5;
+	if ( bfsk_nstopbits < 0 )
+	    bfsk_nstopbits = 1.5;
+    } else if ( strncasecmp(modem_mode, "same",5)==0 ) {
+	// http://www.nws.noaa.gov/nwr/nwrsame.htm
+	bfsk_data_rate = 520.0 + 5/6.0;
+	bfsk_n_data_bits = 8;
+	bfsk_nstartbits = 0;
+	bfsk_nstopbits = 0;
+	bfsk_sync_on_data = 1;
+	bfsk_sync_byte = 0xAB;
+	bfsk_mark_f = 2083.0 + 1/3.0;
+	bfsk_space_f = 1562.5;
+	band_width = bfsk_data_rate;
+    } else if ( strncasecmp(modem_mode, "caller",6)==0 ) {
+	if ( TX_mode ) {
+	    fprintf(stderr, "E: callerid --tx mode is not supported.\n");
+	    return 1;
+	}
+	bfsk_databits_decode = databits_decode_callerid;
+	bfsk_data_rate = 1200;
+	bfsk_n_data_bits = 8;
     } else {
 	bfsk_data_rate = atof(modem_mode);
 	if ( bfsk_n_data_bits == 0 )
@@ -571,15 +631,9 @@ main( int argc, char*argv[] )
     if ( bfsk_data_rate == 0.0 )
 	usage();
 
-    if ( bfsk_n_data_bits == 8 ) {
-	bfsk_framebits_decode = framebits_decode_ascii8;
-	bfsk_framebits_encode = framebits_encode_ascii8;
-    } else if ( bfsk_n_data_bits == 5 ) {
-	bfsk_framebits_decode = framebits_decode_baudot;
-	bfsk_framebits_encode = framebits_encode_baudot;
-    } else {
-	assert( 0 && bfsk_n_data_bits );
-    }
+
+    if ( output_mode_binary )
+	bfsk_databits_decode = databits_decode_binary;
 
     if ( bfsk_data_rate >= 400 ) {
 	/*
@@ -618,8 +672,17 @@ main( int argc, char*argv[] )
 	}
     }
 
-    if ( bfsk_txstopbits == 0 )
-	bfsk_txstopbits = 1.0;
+    // defaults: 1 start bit, 1 stop bit
+    if ( bfsk_nstartbits < 0 )
+	bfsk_nstartbits = 1;
+    if ( bfsk_nstopbits < 0 )
+	bfsk_nstopbits = 1.0;
+
+    if ( bfsk_inverted_freqs ) {
+	float t = bfsk_mark_f;
+	bfsk_mark_f = bfsk_space_f;
+	bfsk_space_f = t;
+    }
 
     /* restrict band_width to <= data rate (FIXME?) */
     if ( band_width > bfsk_data_rate )
@@ -629,7 +692,7 @@ main( int argc, char*argv[] )
     if ( fsk_confidence_search_limit < fsk_confidence_threshold )
 	fsk_confidence_search_limit = fsk_confidence_threshold;
 
-    char *stream_name = NULL;;
+    char *stream_name = NULL;
 
     if ( filename ) {
 	sa_backend = SA_BACKEND_FILE;
@@ -641,7 +704,7 @@ main( int argc, char*argv[] )
      */
     if ( TX_mode ) {
 
-	simpleaudio_tone_init(tx_sin_table_len);
+	simpleaudio_tone_init(tx_sin_table_len, tx_amplitude);
 
 	int tx_interactive = 0;
 	if ( ! stream_name ) {
@@ -650,7 +713,8 @@ main( int argc, char*argv[] )
 	}
 
 	simpleaudio *sa_out;
-	sa_out = simpleaudio_open_stream(sa_backend, SA_STREAM_PLAYBACK,
+	sa_out = simpleaudio_open_stream(sa_backend, sa_backend_device,
+					SA_STREAM_PLAYBACK,
 					sample_format, sample_rate, nchannels,
 					program_name, stream_name);
 	if ( ! sa_out )
@@ -660,8 +724,8 @@ main( int argc, char*argv[] )
 				bfsk_data_rate,
 				bfsk_mark_f, bfsk_space_f,
 				bfsk_n_data_bits,
-				bfsk_txstopbits,
-				bfsk_framebits_encode
+				bfsk_nstopbits,
+				bfsk_databits_encode
 				);
 
 	simpleaudio_close(sa_out);
@@ -678,12 +742,17 @@ main( int argc, char*argv[] )
 	stream_name = "input audio";
 
     simpleaudio *sa;
-    sa = simpleaudio_open_stream(sa_backend, SA_STREAM_RECORD,
+    sa = simpleaudio_open_stream(sa_backend, sa_backend_device,
+				SA_STREAM_RECORD,
 				sample_format, sample_rate, nchannels,
 				program_name, stream_name);
     if ( ! sa )
         return 1;
 
+    sample_rate = simpleaudio_get_rate(sa);
+
+    if ( rxnoise_factor != 0.0 )
+	simpleaudio_set_rxnoise(sa, rxnoise_factor);
 
     /*
      * Prepare the input sample chunk rate
@@ -696,8 +765,7 @@ main( int argc, char*argv[] )
      */
 
     fsk_plan *fskp;
-    fskp = fsk_plan_new(sample_rate, bfsk_mark_f, bfsk_space_f,
-				band_width, bfsk_n_data_bits);
+    fskp = fsk_plan_new(sample_rate, bfsk_mark_f, bfsk_space_f, band_width);
     if ( !fskp ) {
         fprintf(stderr, "fsk_plan_new() failed\n");
         return 1;
@@ -708,14 +776,16 @@ main( int argc, char*argv[] )
      * we need 11 data-bits worth of samples, and we will scan through one bits
      * worth at a time, hence we need a minimum total input buffer size of 12
      * data-bits.  */
-// FIXME I should be able to reduce this to * 9 for 5-bit data, but
-// it SOMETIMES crashes -- probably due to non-integer nsamples_per_bit
-// FIXME by passing it down into the fsk code?
     // FIXME EXPLAIN +1 goes with extra bit when scanning
     size_t	samplebuf_size = ceilf(nsamples_per_bit) * (12+1);
+    samplebuf_size *= 2; // account for the half-buf filling method
+#define SAMPLE_BUF_DIVISOR 12
+#ifdef SAMPLE_BUF_DIVISOR
+    // For performance, use a larger samplebuf_size than necessary
+    if ( samplebuf_size < sample_rate / SAMPLE_BUF_DIVISOR )
+	samplebuf_size = sample_rate / SAMPLE_BUF_DIVISOR;
+#endif
     float	*samplebuf = malloc(samplebuf_size * sizeof(float));
-    float	*samples_readptr = samplebuf;
-    size_t	read_nsamples = samplebuf_size;
     size_t	samples_nvalid = 0;
     debug_log("samplebuf_size=%zu\n", samplebuf_size);
 
@@ -727,6 +797,7 @@ main( int argc, char*argv[] )
 
     int			carrier = 0;
     float		confidence_total = 0;
+    float		amplitude_total = 0;
     unsigned int	nframes_decoded = 0;
     size_t		carrier_nsamples = 0;
 
@@ -742,6 +813,9 @@ main( int argc, char*argv[] )
     // for optimal analysis:
     //   should be >= 0.5 (half a bit width) or we may not find the optimal bit
     //   should be <  1.0 (a full bit width) or we may skip over whole bits
+    // for encodings without start/stop bits:
+    //     MUST be <= 0.5 or we may accidentally skip a bit
+    //
     assert( fsk_frame_overscan >= 0.0 && fsk_frame_overscan < 1.0 );
 
     // ensure that we overscan at least a single sample
@@ -752,16 +826,25 @@ main( int argc, char*argv[] )
     debug_log("fsk_frame_overscan=%f nsamples_overscan=%u\n",
 	    fsk_frame_overscan, nsamples_overscan);
 
+    // n databits plus bfsk_startbit start bits plus bfsk_nstopbit stop bits:
+    float frame_n_bits = bfsk_n_data_bits + bfsk_nstartbits + bfsk_nstopbits;
+    unsigned int frame_nsamples = nsamples_per_bit * frame_n_bits + 0.5;
+
+    float track_amplitude = 0.0;
+
+    signal(SIGINT, rx_stop_sighandler);
+
     while ( 1 ) {
 
+	if ( rx_stop )
+	    break;
+
 	debug_log("advance=%u\n", advance);
 
 	/* Shift the samples in samplebuf by 'advance' samples */
 	assert( advance <= samplebuf_size );
 	if ( advance == samplebuf_size ) {
 	    samples_nvalid = 0;
-	    samples_readptr = samplebuf;
-	    read_nsamples = samplebuf_size;
 	    advance = 0;
 	}
 	if ( advance ) {
@@ -770,24 +853,25 @@ main( int argc, char*argv[] )
 	    memmove(samplebuf, samplebuf+advance,
 		    (samplebuf_size-advance)*sizeof(float));
 	    samples_nvalid -= advance;
-	    samples_readptr = samplebuf + (samplebuf_size-advance);
-	    read_nsamples = advance;
 	}
 
-	/* Read more samples into samplebuf (fill it) */
-	assert ( read_nsamples > 0 );
-	assert ( samples_nvalid + read_nsamples <= samplebuf_size );
-	ssize_t r;
-	r = simpleaudio_read(sa, samples_readptr, read_nsamples);
-	debug_log("simpleaudio_read(samplebuf+%zd, n=%zu) returns %zd\n",
-		samples_readptr - samplebuf, read_nsamples, r);
-	if ( r < 0 ) {
-	    fprintf(stderr, "simpleaudio_read: error\n");
-	    ret = -1;
-            break;
-	}
-	else if ( r > 0 )
+	if ( samples_nvalid < samplebuf_size/2 ) {
+	    float	*samples_readptr = samplebuf + samples_nvalid;
+	    size_t	read_nsamples = samplebuf_size/2;
+	    /* Read more samples into samplebuf (fill it) */
+	    assert ( read_nsamples > 0 );
+	    assert ( samples_nvalid + read_nsamples <= samplebuf_size );
+	    ssize_t r;
+	    r = simpleaudio_read(sa, samples_readptr, read_nsamples);
+	    debug_log("simpleaudio_read(samplebuf+%zd, n=%zu) returns %zd\n",
+		    samples_readptr - samplebuf, read_nsamples, r);
+	    if ( r < 0 ) {
+		fprintf(stderr, "simpleaudio_read: error\n");
+		ret = -1;
+		break;
+	    }
 	    samples_nvalid += r;
+	}
 
 	if ( samples_nvalid == 0 )
 	    break;
@@ -815,13 +899,16 @@ main( int argc, char*argv[] )
 		continue;
 	    }
 
-	    // FIXME: hardcoded negative shift
+	    // default negative shift -- reasonable?
 	    int b_shift = - (float)(autodetect_shift + fskp->band_width/2.0)
 						/ fskp->band_width;
+	    if ( bfsk_inverted_freqs )
+		b_shift *= -1;
 	    /* only accept a carrier as b_mark if it will not result
 	     * in a b_space band which is "too low". */
-	    if ( carrier_band + b_shift < 1 ) {
-		debug_log("autodetected space band too low\n" );
+	    int b_space = carrier_band + b_shift;
+	    if ( b_space < 1 || b_space >= fskp->nbands ) {
+		debug_log("autodetected space band out of range\n" );
 		carrier_band = -1;
 		continue;
 	    }
@@ -839,58 +926,99 @@ main( int argc, char*argv[] )
 
 	debug_log( "--------------------------\n");
 
-	unsigned int frame_nsamples = nsamples_per_bit * fskp->n_frame_bits;
+	// example expect_bits_string
+	//	  0123456789A
+	//	  isddddddddp	i == idle bit (a.k.a. prev_stop bit)
+	//			s == start bit  d == data bits  p == stop bit
+	// ebs = "10dddddddd1"  <-- expected mark/space framing pattern
+	//
+	// NOTE! expect_n_bits ends up being (frame_n_bits+1), because
+	// we expect the prev_stop bit in addition to this frame's own
+	// (start + n_data_bits + stop) bits.  But for each decoded frame,
+	// we will advance just frame_n_bits worth of samples, leaving us
+	// pointing at our stop bit -- it becomes the next frame's prev_stop.
+	//
+	//                  prev_stop--v
+	//                       start--v        v--stop
+	// char *expect_bits_string = "10dddddddd1";
+	//
+	char expect_bits_string[33];
+	int j = 0;
+	if ( bfsk_nstopbits != 0.0 )
+	    expect_bits_string[j++] = '1';
+	if ( bfsk_nstartbits != 0 )	// FIXME -- sets only one start bit
+	    expect_bits_string[j++] = '0';
+	int i;
+	for ( i=0; i<bfsk_n_data_bits; i++,j++ ) {
+	    if ( ! carrier && bfsk_sync_on_data )
+		expect_bits_string[j] = ( (bfsk_sync_byte>>i)&1 ) + '0';
+	    else
+		expect_bits_string[j] = 'd';
+	}
+	if ( bfsk_nstopbits != 0.0 )
+	    expect_bits_string[j++] = '1';
+	expect_bits_string[j++] = 0;
 
-	if ( samples_nvalid < frame_nsamples )
+
+	unsigned int expect_n_bits  = strlen(expect_bits_string);
+	unsigned int expect_nsamples = nsamples_per_bit * expect_n_bits;
+
+	if ( samples_nvalid < expect_nsamples )
 	    break;
 
-	// try_max_nsamples = nsamples_per_bit + nsamples_overscan;
+	// try_max_nsamples
 	// serves two purposes
 	// 1. avoids finding a non-optimal first frame
 	// 2. allows us to track slightly slow signals
-	unsigned int try_max_nsamples = nsamples_per_bit + nsamples_overscan;
+	unsigned int try_max_nsamples;
+	if ( carrier )
+	    try_max_nsamples = nsamples_per_bit * 0.75 + 0.5;
+	else
+	    try_max_nsamples = nsamples_per_bit;
+	try_max_nsamples += nsamples_overscan;
 
-#define FSK_ANALYZE_NSTEPS		10	/* accuracy vs. performance */
-		// Note: FSK_ANALYZE_NSTEPS has subtle effects on the
-		// "rate perfect" calculation.  oh well.
-	unsigned int try_step_nsamples = nsamples_per_bit / FSK_ANALYZE_NSTEPS;
+	// FSK_ANALYZE_NSTEPS Try 4 frame positions across the try_max_nsamples
+	// range.  Using a larger nsteps allows for more accurate tracking of
+	// fast/slow signals (at decreased performance).  Note also
+	// FSK_ANALYZE_NSTEPS_CARRIER_LOCK below, which refines the frame
+	// position upon first acquiring carrier.
+#define FSK_ANALYZE_NSTEPS		3
+	unsigned int try_step_nsamples = try_max_nsamples / FSK_ANALYZE_NSTEPS;
 	if ( try_step_nsamples == 0 )
 	    try_step_nsamples = 1;
 
-	float confidence;
+	float confidence, amplitude;
 	unsigned int bits = 0;
 	/* Note: frame_start_sample is actually the sample where the
 	 * prev_stop bit begins (since the "frame" includes the prev_stop). */
 	unsigned int frame_start_sample = 0;
 
-	// If we don't have carrier, then set this try_confidence_search_limit
-	// to infinity (search for best possible frame) so to get the decoder
-	// into phase with the signal, so the next try_first_sample will match
-	// up with where the next frame should be.
 	unsigned int try_first_sample;
 	float try_confidence_search_limit;
-	if ( carrier ) {
-	    try_first_sample = nsamples_overscan;
-	    try_confidence_search_limit = fsk_confidence_search_limit;
-	} else {
-	    try_first_sample = 0;
-	    try_confidence_search_limit = INFINITY;
-	}
 
-	confidence = fsk_find_frame(fskp, samplebuf, frame_nsamples,
+	try_confidence_search_limit = fsk_confidence_search_limit;
+	try_first_sample = carrier ? nsamples_overscan : 0;
+
+	confidence = fsk_find_frame(fskp, samplebuf, expect_nsamples,
 			try_first_sample,
 			try_max_nsamples,
 			try_step_nsamples,
 			try_confidence_search_limit,
+			expect_bits_string,
 			&bits,
+			&amplitude,
 			&frame_start_sample
 			);
 
-	// FIXME: hardcoded chop off framing bits
-	if ( fskp->n_data_bits == 5 )
-	    bits = ( bits >> 2 ) & 0x1F;
-	else
-	    bits = ( bits >> 2 ) & 0xFF;
+	// no-confidence if amplitude drops abruptly to < 25% of the
+	// track_amplitude, which follows amplitude with hysteresis
+	if ( amplitude < track_amplitude * 0.25 ) {
+	    confidence = 0;
+	} else {
+	    track_amplitude = ( track_amplitude + amplitude ) / 2;
+	    debug_log("@ confidence=%.3f amplitude=%.3f track_amplitude=%.3f\n",
+		    confidence, amplitude, track_amplitude );
+	}
 
 #define FSK_MAX_NOCONFIDENCE_BITS	20
 
@@ -903,12 +1031,17 @@ main( int argc, char*argv[] )
 		if ( carrier ) {
 		    if ( !quiet_mode )
 			report_no_carrier(fskp, sample_rate, bfsk_data_rate,
-			    nsamples_per_bit, nframes_decoded,
-			    carrier_nsamples, confidence_total);
+			    frame_n_bits, nframes_decoded,
+			    carrier_nsamples, confidence_total, amplitude_total);
 		    carrier = 0;
 		    carrier_nsamples = 0;
 		    confidence_total = 0;
+		    amplitude_total = 0;
 		    nframes_decoded = 0;
+		    track_amplitude = 0.0;
+
+		    if ( rx_one )
+			break;
 		}
 	    }
 
@@ -921,7 +1054,7 @@ main( int argc, char*argv[] )
 	}
 
 	// Add a frame's worth of samples to the sample count
-	carrier_nsamples += nsamples_per_bit * (fskp->n_data_bits + 2);
+	carrier_nsamples += frame_nsamples;
 
 	if ( carrier ) {
 
@@ -930,76 +1063,131 @@ main( int argc, char*argv[] )
 	    carrier_nsamples -= nsamples_overscan;
 
 	} else {
+
+	    // We just acquired carrier.
+
 	    if ( !quiet_mode ) {
 		if ( bfsk_data_rate >= 100 )
-		    fprintf(stderr, "### CARRIER %u @ %.1f Hz ###\n",
+		    fprintf(stderr, "### CARRIER %u @ %.1f Hz ",
 			    (unsigned int)(bfsk_data_rate + 0.5),
 			    fskp->b_mark * fskp->band_width);
 		else
-		    fprintf(stderr, "### CARRIER %.2f @ %.1f Hz ###\n",
+		    fprintf(stderr, "### CARRIER %.2f @ %.1f Hz ",
 			    bfsk_data_rate,
 			    fskp->b_mark * fskp->band_width);
 	    }
+
+	    if ( confidence < INFINITY && try_step_nsamples > 1 ) {
+		// FSK_ANALYZE_NSTEPS_CARRIER_LOCK:
+		// Scan again, but try harder to find the best frame upon
+		// acquiring carrier.  Since we found a valid confidence frame
+		// in the "sloppy" fsk_find_frame() call already, we're sure
+		// to find one at least as good this time.
+#define FSK_ANALYZE_NSTEPS_CARRIER_LOCK		8
+		try_step_nsamples = try_max_nsamples / FSK_ANALYZE_NSTEPS_CARRIER_LOCK;
+		if ( try_step_nsamples == 0 )
+		    try_step_nsamples = 1;
+		try_confidence_search_limit = INFINITY;
+		float confidence2, amplitude2;
+		unsigned int bits2;
+		unsigned int frame_start_sample2;
+		confidence2 = fsk_find_frame(fskp, samplebuf, expect_nsamples,
+			    try_first_sample,
+			    try_max_nsamples,
+			    try_step_nsamples,
+			    try_confidence_search_limit,
+			    expect_bits_string,
+			    &bits2,
+			    &amplitude2,
+			    &frame_start_sample2
+			    );
+		if ( confidence2 > confidence ) {
+		    bits = bits2;
+		    amplitude = amplitude2;
+		    frame_start_sample = frame_start_sample2;
+		}
+	    }
+
+	    if ( !quiet_mode )
+		fprintf(stderr, "###\n");
+
 	    carrier = 1;
-	    bfsk_framebits_decode(0, 0, 0);	/* reset the frame processor */
+	    bfsk_databits_decode(0, 0, 0, 0); // reset the frame processor
 	}
 
-
 	confidence_total += confidence;
+	amplitude_total += amplitude;
 	nframes_decoded++;
 	noconfidence = 0;
 
-	/* Advance the sample stream forward past the decoded frame
-	 * but not past the stop bit, since we want it to appear as
-	 * the prev_stop bit of the next frame, so ...
-	 *
-	 * advance = 1 prev_stop + 1 start + N data bits == n_data_bits+2
-	 *
-	 * but actually advance just a bit less than that to allow
-	 * for tracking slightly fast signals, hence - nsamples_overscan.
-	 */
-	advance = frame_start_sample
-		+ nsamples_per_bit * (float)(fskp->n_data_bits + 2)
-		- nsamples_overscan;
+	// Advance the sample stream forward past the junk before the
+	// frame starts (frame_start_sample), and then past decoded frame
+	// (see also NOTE about frame_n_bits and expect_n_bits)...
+	// But actually advance just a bit less than that to allow
+	// for tracking slightly fast signals, hence - nsamples_overscan.
+	advance = frame_start_sample + frame_nsamples - nsamples_overscan;
 
 	debug_log("@ nsamples_per_bit=%.3f n_data_bits=%u "
 			" frame_start=%u advance=%u\n",
-		    nsamples_per_bit, fskp->n_data_bits,
+		    nsamples_per_bit, bfsk_n_data_bits,
 		    frame_start_sample, advance);
 
+	// chop off the prev_stop bit
+	if ( bfsk_nstopbits != 0.0 )
+	    bits = bits >> 1;
+
 
 	/*
 	 * Send the raw data frame bits to the backend frame processor
 	 * for final conversion to output data bytes.
 	 */
 
+	// chop off framing bits
+	unsigned int frame_bits_shift = bfsk_nstartbits;
+	unsigned int frame_bits_mask = (int)(1<<bfsk_n_data_bits) - 1;
+	bits = ( bits >> frame_bits_shift ) & frame_bits_mask;
+
 	unsigned int dataout_size = 4096;
 	char dataoutbuf[4096];
 	unsigned int dataout_nbytes = 0;
 
-	dataout_nbytes += bfsk_framebits_decode(dataoutbuf + dataout_nbytes,
+	// suppress printing of bfsk_sync_byte bytes
+	if ( bfsk_sync_on_data ) {
+	    if ( dataout_nbytes == 0 && bits == bfsk_sync_byte )
+		continue;
+	}
+
+	dataout_nbytes += bfsk_databits_decode(dataoutbuf + dataout_nbytes,
 						dataout_size - dataout_nbytes,
-						bits);
+						bits, (int)bfsk_n_data_bits);
+
+	if ( dataout_nbytes == 0 )
+	    continue;
 
 	/*
 	 * Print the output buffer to stdout
 	 */
-	if ( dataout_nbytes ) {
+	if ( output_print_filter == 0 ) {
+	    if ( write(1, dataoutbuf, dataout_nbytes) < 0 )
+		perror("write");
+	} else {
 	    char *p = dataoutbuf;
 	    for ( ; dataout_nbytes; p++,dataout_nbytes-- ) {
 		char printable_char = isprint(*p)||isspace(*p) ? *p : '.';
-		printf( "%c", printable_char );
+		if ( write(1, &printable_char, 1) < 0 )
+		    perror("write");
 	    }
-	    fflush(stdout);
 	}
 
     } /* end of the main loop */
 
+    signal(SIGINT, SIG_DFL);
+
     if ( carrier ) {
 	if ( !quiet_mode )
 	    report_no_carrier(fskp, sample_rate, bfsk_data_rate,
-		nsamples_per_bit, nframes_decoded,
-		carrier_nsamples, confidence_total);
+		frame_n_bits, nframes_decoded,
+		carrier_nsamples, confidence_total, amplitude_total);
     }
 
     simpleaudio_close(sa);
